//==============================================================================
//
//  Provider Base Class 
//
//  Created by Kwon Keuk Han
//  Copyright (c) 2018 AirenSoft. All rights reserved.
//
//==============================================================================

#include "provider.h"
#include "application.h"
#include "stream.h"
#include "provider_private.h"

#include "orchestrator/orchestrator.h"
namespace pvd
{
	Application::Application(const std::shared_ptr<Provider> &provider, const info::Application &application_info)
		: info::Application(application_info)
	{
		_provider = provider;
	}

	Application::~Application()
	{
		Stop();
	}

	const char* Application::GetApplicationTypeName()
	{
		if(_provider == nullptr)
		{
			return "";
		}

		if(_app_type_name.IsEmpty())
		{
			_app_type_name.Format("%s %s",  _provider->GetProviderName(), "Application");
		}

		return _app_type_name.CStr();
	}

	bool Application::Start()
	{
		logti("%s has created [%s] application", _provider->GetProviderName(), GetName().CStr());

		_state = ApplicationState::Started;

		return true;
	}

	bool Application::Stop()
	{
		if(_state == ApplicationState::Stopped)
		{
			return true;
		}

		DeleteAllStreams();
		logti("%s has deleted [%s] application", _provider->GetProviderName(), GetName().CStr());
		_state = ApplicationState::Stopped;
		return true;
	}

	info::stream_id_t Application::IssueUniqueStreamId()
	{
		static std::atomic<info::stream_id_t>	last_issued_stream_id(100);

		return last_issued_stream_id++;
	}

	const std::shared_ptr<Stream> Application::GetStreamById(uint32_t stream_id)
	{
		std::shared_lock<std::shared_mutex> lock(_streams_guard);

		if(_streams.find(stream_id) == _streams.end())
		{
			return nullptr;
		}

		return _streams.at(stream_id);
	}

	const std::shared_ptr<Stream> Application::GetStreamByName(ov::String stream_name)
	{
		std::shared_lock<std::shared_mutex> lock(_streams_guard);
		
		for(auto const &x : _streams)
		{
			auto& stream = x.second;
			if(stream->GetName() == stream_name)
			{
				return stream;
			}
		}

		return nullptr;
	}

	bool Application::AddStream(const std::shared_ptr<Stream> &stream)
	{
		// Check if same stream name is exist in MediaRouter(may be created by another provider)
		if (IsExistingInboundStream(stream->GetName()) == true)
		{
			logtw("Reject to add stream : there is already an incoming stream (%s) with the same name in application(%s) ", stream->GetName().CStr(), GetName().CStr());
			return false;
		}

		// If provider is OVT, it is running in Edge mode.
		if (_provider->GetProviderType() != ProviderType::Ovt)
		{
			// Register stream if OriginMapStore is enabled
			auto result = ocst::Orchestrator::GetInstance()->RegisterStreamToOriginMapStore(GetName(), stream->GetName());
			if (result == CommonErrorCode::ERROR)
			{
				logtw("Reject to add stream : failed to register stream to origin map store");
				return false;
			}
		}

		// If there is no data track, add data track
		if (stream->GetFirstTrack(cmn::MediaType::Data) == nullptr)
		{
			// Add data track 
			auto data_track = std::make_shared<MediaTrack>();

			// Issue unique track id
			data_track->SetId(stream->IssueUniqueTrackId());
			data_track->SetMediaType(cmn::MediaType::Data);
			data_track->SetTimeBase(1, 1000);
			data_track->SetOriginBitstream(cmn::BitstreamFormat::Unknown);
			
			stream->AddTrack(data_track);
		}

		stream->SetApplication(GetSharedPtrAs<Application>());
		stream->SetApplicationInfo(GetSharedPtrAs<Application>());

		// This is not an official feature
		// OutputProfile(without encoding) is not applied to a specific provider.
		// To reduce transcoding cost when using Persistent Stream
		if (_provider->GetProviderType() == ProviderType::Rtmp)
		{
			if(GetConfig().GetProviders().GetRtmpProvider().IsPassthroughOutputProfile() == true)
			{
				stream->SetRepresentationType(StreamRepresentationType::Relay);
			}
		}

		std::unique_lock<std::shared_mutex> streams_lock(_streams_guard);
		_streams[stream->GetId()] = stream;
		streams_lock.unlock();

		NotifyStreamCreated(stream);
		
		return true;
	}

	bool Application::DeleteStream(const std::shared_ptr<Stream> &stream)
	{
		std::unique_lock<std::shared_mutex> streams_lock(_streams_guard);

		if(_streams.find(stream->GetId()) == _streams.end())
		{
			logtc("Could not find stream to be removed : %s/%s(%u)", stream->GetApplicationInfo().GetName().CStr(), stream->GetName().CStr(), stream->GetId());
			return false;
		}
		_streams.erase(stream->GetId());

		streams_lock.unlock();
		
		stream->Stop();

		NotifyStreamDeleted(stream);

		// If provider is OVT, it is running in Edge mode.
		if (_provider->GetProviderType() != ProviderType::Ovt)
		{
			// Unegister stream if OriginMapStore is enabled
			auto result = ocst::Orchestrator::GetInstance()->UnregisterStreamFromOriginMapStore(GetName(), stream->GetName());
			if (result == CommonErrorCode::ERROR)
			{
				logtw("Reject to add stream : failed to register stream to origin map store");
				return false;
			}
		}

		return true;
	}

	bool Application::NotifyStreamCreated(const std::shared_ptr<Stream> &stream)
	{
		return MediaRouteApplicationConnector::CreateStream(stream);
	}

	bool Application::NotifyStreamUpdated(const std::shared_ptr<info::Stream> &stream)
	{
		return MediaRouteApplicationConnector::UpdateStream(stream);
	}

	bool Application::NotifyStreamDeleted(const std::shared_ptr<Stream> &stream)
	{
		return MediaRouteApplicationConnector::DeleteStream(stream);
	}

	bool Application::DeleteAllStreams()
	{
		std::unique_lock<std::shared_mutex> lock(_streams_guard);

		for(auto it = _streams.cbegin(); it != _streams.cend(); )
		{
			auto stream = it->second;
			it = _streams.erase(it);
			stream->Stop();

			NotifyStreamDeleted(stream);
		}

		return true;
	}
}