//==============================================================================
//
//  OvenMediaEngine
//
//  Created by Getroot
//  Copyright (c) 2022 AirenSoft. All rights reserved.
//
//==============================================================================
#include "llhls_master_playlist.h"
#include "llhls_private.h"

#include <base/ovlibrary/zip.h>
#include <modules/bitstream/codec_media_type.h>

void LLHlsMasterPlaylist::SetChunkPath(const ov::String &chunk_path)
{
	_chunk_path = chunk_path;
}

// Add X-MEDIA to the master playlist
void LLHlsMasterPlaylist::AddMediaCandidateToMasterPlaylist(const ov::String &group_id, const std::shared_ptr<const MediaTrack> &track, const ov::String &chunk_uri)
{
	LLHlsMasterPlaylist::MediaInfo media_info;

	media_info._type = track->GetMediaType()==cmn::MediaType::Video ? LLHlsMasterPlaylist::MediaInfo::Type::Video : LLHlsMasterPlaylist::MediaInfo::Type::Audio;
	media_info._group_id = group_id; // Currently there is only one element per group.
	media_info._name = "none";
	media_info._default = true; // Currently there is no group media so all track is default
	media_info._uri = chunk_uri;
	media_info._track = track;

	//TODO(Getroot) : If a group such as multilingual audio is supported in the future, the language value of the track must be supported.

	AddGroupMedia(media_info);
}

// Add X-STREAM-INF to the master playlist
void LLHlsMasterPlaylist::AddStreamInfToMasterPlaylist(const std::shared_ptr<const MediaTrack> &video_track, const ov::String &video_chunk_uri, 
														const std::shared_ptr<const MediaTrack> &audio_track, const ov::String &audio_chunk_uri)
{
	LLHlsMasterPlaylist::StreamInfo stream_info;
	if (video_track != nullptr)
	{
		stream_info._track = video_track;
		stream_info._uri = video_chunk_uri;
		
		if (audio_track != nullptr)
		{
			stream_info._audio_group_id = ov::Converter::ToString(audio_track->GetId());
		}
	}
	else if (audio_track != nullptr)
	{
		stream_info._track = audio_track;
		stream_info._uri = audio_chunk_uri;
	}

	AddStreamInfo(stream_info);
}


bool LLHlsMasterPlaylist::AddGroupMedia(const MediaInfo &media_info)
{
	if (media_info._group_id.IsEmpty())
	{
		return false;
	}

	// Validate media info
	// TODO(Getroot) : It it better to validate media info in the Config Module
	auto it = _media_infos.find(media_info._group_id);
	if (it != _media_infos.end())
	{
		auto first = it->second.front();
		if (first._type != media_info._type || 
			first._track->GetCodecId() != media_info._track->GetCodecId())
		{
			return false;
		}
	}

	// lock
	std::lock_guard<std::shared_mutex> lock(_media_infos_guard);
	_media_infos[media_info._group_id].push_back(media_info);

	return true;
}

// Video + Audio + Subtitle + ClosedCaptions
// But now OME support only Video + Audio
bool LLHlsMasterPlaylist::AddStreamInfo(const StreamInfo &stream_info)
{
	if (stream_info._uri.IsEmpty())
	{
		return false;
	}

	// Video group
	if (!stream_info._video_group_id.IsEmpty())
	{
		SetActiveMediaGroup(stream_info._video_group_id);
	}

	// Audio group
	if (!stream_info._audio_group_id.IsEmpty())
	{
		SetActiveMediaGroup(stream_info._audio_group_id);
	}

	// Subtitle group
	if (!stream_info._subtitle_group_id.IsEmpty())
	{
		SetActiveMediaGroup(stream_info._subtitle_group_id);
	}

	// Closed captions group
	if (!stream_info._closed_captions_group_id.IsEmpty())
	{
		SetActiveMediaGroup(stream_info._closed_captions_group_id);
	}
	
	// lock
	std::lock_guard<std::shared_mutex> lock(_stream_infos_guard);
	_stream_infos.push_back(stream_info);

	return true;
}

bool LLHlsMasterPlaylist::SetActiveMediaGroup(const ov::String &group_id)
{
	if (group_id.IsEmpty() == false)
	{
		std::shared_lock<std::shared_mutex> media_infos_lock(_media_infos_guard);
		auto it = _media_infos.find(group_id);
		if (it == _media_infos.end())
		{
			return false;
		}

		auto &media_infos = it->second;
		for (auto &media_info : media_infos)
		{
			media_info._used = true;
		}
	}

	return true;
}

const LLHlsMasterPlaylist::MediaInfo &LLHlsMasterPlaylist::GetDefaultMediaInfo(const ov::String &group_id) const
{
	std::shared_lock<std::shared_mutex> media_infos_lock(_media_infos_guard);

	auto it = _media_infos.find(group_id);
	if (it == _media_infos.end())
	{
		return MediaInfo::InvalidMediaInfo();
	}

	auto &media_infos = it->second;
	for (auto &media_info : media_infos)
	{
		if (media_info._default == true)
		{
			return media_info;
		}
	}

	return MediaInfo::InvalidMediaInfo();
}

void LLHlsMasterPlaylist::UpdateCacheForDefaultPlaylist()
{
	ov::String playlist = MakePlaylist("", false);
	{
		// lock 
		std::lock_guard<std::shared_mutex> lock(_cached_default_playlist_guard);
		_cached_default_playlist = playlist;
	}

	{
		// lock 
		std::lock_guard<std::shared_mutex> lock(_cached_default_playlist_gzip_guard);
		_cached_default_playlist_gzip = ov::Zip::CompressGzip(playlist.ToData(false));
	}
}

ov::String LLHlsMasterPlaylist::MakePlaylist(const ov::String &chunk_query_string, bool legacy, bool include_path) const
{
	ov::String playlist(10240);

	playlist.AppendFormat("#EXTM3U\n");
	playlist.AppendFormat("#EXT-X-VERSION:7\n");
	playlist.AppendFormat("#EXT-X-INDEPENDENT-SEGMENTS\n");

	// Write EXT-X-MEDIA from _media_infos
	// shared lock
	std::shared_lock<std::shared_mutex> media_infos_lock(_media_infos_guard);
	for (const auto &[group_id, media_infos] : _media_infos)
	{
		for (const auto &media_info : media_infos)
		{
			if (media_info._used)
			{
				playlist.AppendFormat("#EXT-X-MEDIA:TYPE=%s", media_info.TypeString().CStr());
				playlist.AppendFormat(",GROUP-ID=\"%s\"", media_info._group_id.CStr());
				playlist.AppendFormat(",NAME=\"%s\"", media_info._name.CStr());
				playlist.AppendFormat(",AUTOSELECT=%s", media_info._auto_select ? "YES" : "NO");
				playlist.AppendFormat(",DEFAULT=%s", media_info._default ? "YES" : "NO");

				if (media_info._type == MediaInfo::Type::Audio)
				{
					playlist.AppendFormat(",CHANNELS=\"%d\"", media_info._track->GetChannel().GetCounts() == 0 ? 2 : media_info._track->GetChannel().GetCounts());
				}

				if (!media_info._language.IsEmpty())
				{
					playlist.AppendFormat(",LANGUAGE=\"%s\"", media_info._language.CStr());
				}

				if (!media_info._instream_id.IsEmpty())
				{
					playlist.AppendFormat(",INSTREAM-ID=\"%s\"", media_info._instream_id.CStr());
				}

				// _assoc_language
				if (!media_info._assoc_language.IsEmpty())
				{
					playlist.AppendFormat(",ASSOCIATED-LANGUAGE=\"%s\"", media_info._assoc_language.CStr());
				}

				if (!media_info._uri.IsEmpty())
				{
					playlist.AppendFormat(",URI=\"%s%s", include_path?_chunk_path.CStr():"", media_info._uri.CStr());

					if (chunk_query_string.IsEmpty() == false)
					{
						playlist.AppendFormat("?%s", chunk_query_string.CStr());
					}

					if (legacy)
					{
						if (chunk_query_string.IsEmpty() == false)
						{
							playlist.AppendFormat("&_HLS_legacy=YES");
						}
						else
						{
							playlist.AppendFormat("?_HLS_legacy=YES");
						}
					}
				}

				playlist.AppendFormat("\"\n");
			}
		}
	}
	media_infos_lock.unlock();

	playlist.AppendFormat("\n");

	// Write EXT-X-STREAM-INF from _stream_infos
	// shared lock
	std::shared_lock<std::shared_mutex> stream_infos_lock(_stream_infos_guard);
	for (const auto &variant_info : _stream_infos)
	{
		playlist.AppendFormat("#EXT-X-STREAM-INF:");

		std::shared_ptr<const MediaTrack> video_track = nullptr, audio_track = nullptr;

		if (variant_info._track->GetMediaType() == cmn::MediaType::Video)
		{
			video_track = variant_info._track;
			if (variant_info._audio_group_id.IsEmpty() == false)
			{
				auto &audio_media_info = GetDefaultMediaInfo(variant_info._audio_group_id);
				audio_track = audio_media_info._track;
			}
		}
		else if (variant_info._track->GetMediaType() == cmn::MediaType::Audio)
		{
			audio_track = variant_info._track;
			if (variant_info._video_group_id.IsEmpty() == false)
			{
				auto &video_media_info = GetDefaultMediaInfo(variant_info._video_group_id);
				video_track = video_media_info._track;
			}
		}

		// BANDWIDTH
		uint32_t bandwidth = 0;

		if (video_track != nullptr)
		{
			bandwidth += video_track->GetBitrate();
		}
		
		if (audio_track != nullptr)
		{
			bandwidth += audio_track->GetBitrate();
		}

		// Output Bandwidth
		playlist.AppendFormat("BANDWIDTH=%d", bandwidth == 0 ? 2000000 : bandwidth);

		// If variant_info is video, output RESOLUTION, FRAME-RATE
		if (variant_info._track->GetMediaType() == cmn::MediaType::Video)
		{
			playlist.AppendFormat(",RESOLUTION=%dx%d", variant_info._track->GetWidth(), variant_info._track->GetHeight());
			playlist.AppendFormat(",FRAME-RATE=%.1f", variant_info._track->GetFrameRate() == 0 ? 30.0 : variant_info._track->GetFrameRate());
		}

		// CODECS
		playlist.AppendFormat(",CODECS=\"");

		if (video_track != nullptr)
		{
			playlist.AppendFormat("%s", CodecMediaType::GetCodecsParameter(video_track).CStr());
		}
		if (audio_track != nullptr)
		{
			if (video_track != nullptr)
			{
				playlist.AppendFormat(",");
			}
			playlist.AppendFormat("%s", CodecMediaType::GetCodecsParameter(audio_track).CStr());
		}
		playlist.AppendFormat("\"");

		if (variant_info._video_group_id.IsEmpty() == false)
		{
			playlist.AppendFormat(",VIDEO=\"%s\"", variant_info._video_group_id.CStr());
		}

		if (variant_info._audio_group_id.IsEmpty() == false)
		{
			playlist.AppendFormat(",AUDIO=\"%s\"", variant_info._audio_group_id.CStr());
		}

		// URI
		playlist.AppendFormat("\n");
		playlist.AppendFormat("%s%s", include_path?_chunk_path.CStr():"", variant_info._uri.CStr());
		if (chunk_query_string.IsEmpty() == false)
		{
			playlist.AppendFormat("?%s", chunk_query_string.CStr());
		}
		if (legacy)
		{
			if (chunk_query_string.IsEmpty() == false)
			{
				playlist.AppendFormat("&_HLS_legacy=YES");
			}
			else
			{
				playlist.AppendFormat("?_HLS_legacy=YES");
			}
		}
		playlist.AppendFormat("\n");
	}
	stream_infos_lock.unlock();

	return playlist;
}

ov::String LLHlsMasterPlaylist::ToString(const ov::String &chunk_query_string, bool legacy, bool include_path) const
{
	if (chunk_query_string.IsEmpty() && legacy == false && include_path == true && !_cached_default_playlist.IsEmpty())
	{
		std::shared_lock<std::shared_mutex> lock(_cached_default_playlist_guard);
		return _cached_default_playlist;
	}

	return MakePlaylist(chunk_query_string, legacy, include_path);
}

std::shared_ptr<const ov::Data> LLHlsMasterPlaylist::ToGzipData(const ov::String &chunk_query_string, bool legacy) const
{
	if (chunk_query_string.IsEmpty() && legacy == false && _cached_default_playlist_gzip != nullptr)
	{
		std::shared_lock<std::shared_mutex> lock(_cached_default_playlist_gzip_guard);
		return _cached_default_playlist_gzip;
	}

	return  ov::Zip::CompressGzip(ToString(chunk_query_string, legacy).ToData(false));
}